import abc
from typing import List, Any, Optional, AsyncIterator

from suppgram.entities import (
    CustomerIdentification,
    Customer,
    AgentIdentification,
    Agent,
    AgentDiff,
    WorkplaceIdentification,
    Workplace,
    Conversation,
    Message,
    ConversationDiff,
    CustomerDiff,
    Tag,
    Event,
)


class Storage(abc.ABC):
    """
    Storage encapsulates functionality related to storing data persistently in a database,
    allowing to integrate Suppgram into systems with various tech stacks.

    Basically implements Repository design pattern.
    """

    async def initialize(self):
        """Performs asynchronous initialization if needed."""

    @abc.abstractmethod
    async def create_or_update_customer(
        self,
        identification: CustomerIdentification,
        diff: Optional[CustomerDiff] = None,
    ) -> Customer:
        """Creates or updates customer identified by given `identification` with new data provided in `diff`.

        Can be used with `diff=None` to get already existing customer with no changes applied.

        Parameters:
            identification: data necessary to uniquely identify the customer.
            diff: optional metadata to be updated.
        """

    @abc.abstractmethod
    async def get_agent(self, identification: AgentIdentification) -> Agent:
        pass

    @abc.abstractmethod
    def find_all_agents(self) -> AsyncIterator[Agent]:
        pass

    @abc.abstractmethod
    async def create_or_update_agent(
        self, identification: AgentIdentification, diff: Optional[AgentDiff] = None
    ) -> Agent:
        pass

    @abc.abstractmethod
    async def update_agent(self, identification: AgentIdentification, diff: AgentDiff) -> Agent:
        pass

    @abc.abstractmethod
    async def get_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        pass

    @abc.abstractmethod
    async def get_agent_workplaces(self, agent: Agent) -> List[Workplace]:
        pass

    @abc.abstractmethod
    async def get_or_create_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        pass

    @abc.abstractmethod
    async def create_tag(self, name: str, created_by: Agent) -> Tag:
        pass

    @abc.abstractmethod
    async def find_all_tags(self) -> List[Tag]:
        pass

    @abc.abstractmethod
    async def get_or_create_conversation(self, customer: Customer) -> Conversation:
        pass

    @abc.abstractmethod
    async def find_conversations_by_ids(
        self, conversation_ids: List[Any], with_messages: bool = False
    ) -> List[Conversation]:
        pass

    @abc.abstractmethod
    def find_all_conversations(self, with_messages: bool = False) -> AsyncIterator[Conversation]:
        pass

    @abc.abstractmethod
    async def count_all_conversations(self) -> int:
        pass

    @abc.abstractmethod
    async def update_conversation(
        self, id: Any, diff: ConversationDiff, unassigned_only: bool = False
    ):
        pass

    @abc.abstractmethod
    async def get_agent_conversation(self, identification: WorkplaceIdentification) -> Conversation:
        pass

    @abc.abstractmethod
    async def find_customer_conversations(
        self, customer: Customer, with_messages: bool = False
    ) -> List[Conversation]:
        pass

    @abc.abstractmethod
    async def find_agent_conversations(
        self, agent: Agent, with_messages: bool = False
    ) -> List[Conversation]:
        pass

    @abc.abstractmethod
    async def save_message(self, conversation: Conversation, message: Message):
        pass

    @abc.abstractmethod
    async def save_event(self, event: Event):
        pass

    @abc.abstractmethod
    def find_all_events(self) -> AsyncIterator[Event]:
        """Iterate over all stored events in chronological order."""

    @abc.abstractmethod
    async def count_all_events(self) -> int:
        pass



from .storage import TelegramStorage  # noqa
from .agent_frontend import TelegramAgentFrontend  # noqa
from .customer_frontend import TelegramCustomerFrontend  # noqa
from .manager_frontend import TelegramManagerFrontend  # noqa



import asyncio

from pubnub.callbacks import SubscribeCallback
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.pubsub import PNPresenceEventResult, PNMessageResult
from pubnub.pubnub import PubNub

from suppgram.backend import Backend
from suppgram.entities import MessageKind, NewMessageForCustomerEvent
from suppgram.frontend import CustomerFrontend
from suppgram.frontends.pubnub.configuration import Configuration
from suppgram.frontends.pubnub.converter import MessageConverter
from suppgram.frontends.pubnub.identification import make_customer_identification


class PubNubCustomerFrontend(CustomerFrontend):
    """
    Allows customers to access the support system via PubNub channels.

    All channels from a specific PubNub channel group will be considered
    individual chats with support. When agent responds, agent's message content
    will be copied and sent to the customer on behalf of a particular PubNub user
    (e.g. with user ID "support").
    """

    def __init__(
        self,
        backend: Backend,
        message_converter: MessageConverter,
        pubnub_channel_group: str,
        pubnub_configuration: Configuration,  # user ID to send messages from is also in there
    ):
        """
        This constructor should not be used directly; use [Builder][suppgram.builder.Builder] instead.

        Arguments:
            backend: used backend instance.
            message_converter: helper object responsible for converting messages between Suppgram dataclasses
                               and project-specific PubNub JSONs.
            pubnub_channel_group: name of PubNub channel group which includes all individual chats with support.
                                  Will be used to subscribe for updates from all those chats at once.
            pubnub_configuration: PubNub client configuration.
        """
        self._message_converter = message_converter
        self._pubnub = pubnub_configuration.instantiate_async()
        self._pubnub.add_listener(_SubscribeCallback(backend, message_converter))
        self._pubnub_channel_group = pubnub_channel_group
        backend.on_new_message_for_customer.add_handler(self._handle_new_message_for_customer_event)

    async def start(self):
        self._pubnub.subscribe().channel_groups([self._pubnub_channel_group]).execute()

    async def _handle_new_message_for_customer_event(self, event: NewMessageForCustomerEvent):
        customer = event.conversation.customer
        if customer.pubnub_user_id is None or customer.pubnub_channel_id is None:
            return
        converted = self._message_converter.convert_to_pubnub(event.message)
        (
            await self._pubnub.publish()
            .channel(customer.pubnub_channel_id)
            .message(converted)
            .future()
        )


class _SubscribeCallback(SubscribeCallback):
    def __init__(self, backend: Backend, message_converter: MessageConverter):
        self._backend = backend
        self._message_converter = message_converter

    def status(self, pubnub: PubNub, status: PNStatus):
        pass

    def message(self, pubnub: PubNub, message: PNMessageResult):
        loop = asyncio.get_event_loop()
        loop.create_task(self._message(message))

    async def _message(self, message: PNMessageResult):
        converted = self._message_converter.convert_from_pubnub(message)
        if converted.kind != MessageKind.FROM_CUSTOMER:
            return
        identification = make_customer_identification(message)
        conversation = await self._backend.identify_customer_conversation(identification)
        await self._backend.process_message(conversation, converted)

    def presence(self, pubnub: PubNub, presence: PNPresenceEventResult):
        pass



import logging

from suppgram.entities import (
    Conversation,
    Message,
    MessageKind,
    Agent,
)
from suppgram.texts.interface import TextProvider, Text, Format

logger = logging.getLogger(__name__)


class RussianTextProvider(TextProvider):
    telegram_customer_unsupported_message_content = (
        "😞 Извините, такой тип сообщений пока не поддерживается. "
        "Агент поддержки не увидит это сообщение."
    )
    telegram_agent_unsupported_message_content = (
        "😞 Извините, такой тип сообщений пока не поддерживается. "
        "Пользователь не увидит это сообщение."
    )

    telegram_customer_start_message = (
        "👋 Добро пожаловать в службу поддержки! Опишите свою проблему."
    )
    telegram_customer_conversation_resolved_message_placeholder = (
        "✅ Обсуждение было завершено агентом поддержки. "
        "Вы всегда можете начать новое обсуждение, написав в этот чат!"
    )
    telegram_customer_conversation_resolved_message = (
        telegram_customer_conversation_resolved_message_placeholder + "\n\n"
        "⭐️ Пожалуйста, оцените работу агента:"
    )

    def compose_customer_conversation_resolved_message(self, rating: int) -> str:
        return (
            self.telegram_customer_conversation_resolved_message_placeholder
            + "\n\n⭐️ Ваша оценка работы агента: "
            + self.format_rating(rating)
        )

    telegram_manager_start_message = "🛠️ Добро пожаловать в бота-администратора службы поддержки!"
    telegram_manager_permission_denied_message = "🚫 У вас нет доступа к этой функциональности."
    telegram_agents_command_description = "Сделать всех участников этого чата агентами поддержки."
    telegram_agents_command_success_message = (
        "Теперь я буду считать всех участников этого чата агентами службы поддержки."
    )
    telegram_send_new_conversations_command_description = (
        "Отправлять уведомления о новых обсуждениях в этот чат."
    )
    telegram_send_new_conversations_command_success_message = (
        "Теперь я буду отправлять уведомления о всех новых обсуждениях в этот чат."
    )
    telegram_report_command_description = "Выгрузить отчёт"
    telegram_report_message_placeholder = "⏳ Расчёт статистики... {progress}"
    telegram_report_message = (
        "📊 Отчёт\n\n"
        "Среднее время до первого ответа: {report.average_start_to_first_response_time_min:.1f} мин\n"
        "Среднее время завершения обсуждения: {report.average_start_to_resolution_time_min:.1f} мин\n"
        "Средняя оценка: {report.average_customer_rating:.1f}"
    )

    telegram_create_tag_command_description = "Создать новый тег для обсуждений."
    telegram_create_tag_permission_denied_message = "🚫 У вас нет доступа к созданию тегов."
    telegram_create_tag_usage_message = (
        "🧑‍🏫 Пожалуйста, укажите название нового тега после команды:\n\n    /create_tag urgent"
    )
    telegram_tag_successfully_created_message = "✅ Новый тег создан."
    telegram_tag_already_exists_message = "⚠️ Тег с таким названием уже есть!"

    telegram_agent_start_message = "👷 Добро пожаловать в бота для агентов поддержки!"
    telegram_agent_permission_denied_message = "🚫 У вас нет доступа к этой функциональности."
    telegram_workplace_is_not_assigned_message = (
        "📭 Сейчас этот чат не связан с кем-либо из пользователей."
    )
    telegram_resolve_command_description = "Пометить обсуждение как завершённое."
    telegram_postpone_command_description = (
        "Вернуть обсуждение в статус NEW и перестать переписываться с пользователем."
    )
    telegram_agent_conversation_resolved_message = (
        "✅ Обсуждение было завершено. Этот чат больше не связан с кем-либо из пользователей."
    )
    telegram_new_conversation_notification_placeholder = "❗️ Новое обсуждение!"

    customer_profile_header = "👤 Пользователь: {customer}"
    customer_profile_anonymous = "аноним"
    customer_profile_contacts = "📒 Контакты: {contacts}"
    customer_rating_footer = "🎖 Оценка: {rating}"

    def compose_conversation_notification_header(self, conversation: Conversation) -> str:
        emoji = self.CONVERSATION_STATE_TO_EMOJI.get(conversation.state, "")
        return f"{emoji} Обсуждение в статусе #{conversation.state.upper()}"

    conversation_notification_assigned_to = "Назначено, агент {agent}"

    def compose_nudge_to_start_bot_notification(
        self, agent: Agent, telegram_bot_username: str
    ) -> Text:
        if agent.telegram_user_id is None:
            raise RuntimeError("agent without Telegram account can't /start a bot!")
        agent_ref = self._format_telegram_mention(
            telegram_user_id=agent.telegram_user_id,
            telegram_first_name=agent.telegram_first_name,
            telegram_last_name=None,  # less formal
            telegram_username=agent.telegram_username,
            format_=Format.TELEGRAM_MARKDOWN,
        )
        return Text(
            text=f"⚠️ {agent_ref}, пожалуйста, откройте бота @{telegram_bot_username} и "
            f"нажмите кнопку Start или Restart, чтобы иметь возможность переписываться с пользователем. \\.",
            format=Format.TELEGRAM_MARKDOWN,
        )

    telegram_assign_to_me_button_text = "Назначить на меня"

    message_history_title = "🗂️ История сообщений\n"

    def format_history_message(self, message: Message) -> str:
        if message.kind == MessageKind.FROM_CUSTOMER:
            return f"👤 Пользователь: {self.format_text_or_media(message)}"
        if message.kind == MessageKind.FROM_AGENT:
            return f"🧑‍💼 Агент: {self.format_text_or_media(message)}"
        if message.kind == MessageKind.POSTPONED:
            return "⏳ Обсуждение было отложено."
        if message.kind == MessageKind.RESOLVED:
            return "✅ Обсуждение было завершено."
        logger.warning(f"Unsupported message kind: {message.kind.value!r}")
        return str(message.kind.value)

    def format_text_or_media(self, message: Message) -> str:
        if message.text:
            return message.text
        if message.image:
            return "[🖼️изображение]"
        return f"[{message.media_kind}]"



import abc
from datetime import datetime, timezone
from importlib import import_module
from typing import Any

from pubnub.models.consumer.pubsub import PNMessageResult

from suppgram.entities import Message, MessageKind


def make_pubnub_message_converter(converter_class_path: str) -> "MessageConverter":
    (
        converter_module_name,
        converter_class_name,
    ) = converter_class_path.rsplit(".", 1)
    converter_class = getattr(import_module(converter_module_name), converter_class_name)
    return converter_class()


class MessageConverter(abc.ABC):
    """Converts messages from project-specific JSON to [Message][suppgram.entities.Message] dataclass and vice-versa.

    Methods:
        convert_from_pubnub: convert from project-specific JSON received from PubNub
        convert_to_pubnub: convert to project-specific JSON to be sent to PubNub"""

    @abc.abstractmethod
    def convert_from_pubnub(self, message: PNMessageResult) -> Message:
        pass

    @abc.abstractmethod
    def convert_to_pubnub(self, message: Message) -> Any:
        pass


class DefaultMessageConverter(MessageConverter):
    _ISO_8601_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"

    def convert_from_pubnub(self, message: PNMessageResult) -> Message:
        return Message(
            kind=MessageKind(message.message["kind"]),
            time_utc=datetime.strptime(message.message["time"], self._ISO_8601_FORMAT).astimezone(
                timezone.utc
            ),
            text=message.message.get("text"),
        )

    def convert_to_pubnub(self, message: Message) -> Any:
        return {
            "kind": message.kind.value,
            "time": message.time_utc.strftime(self._ISO_8601_FORMAT),
            "text": message.text,
        }



import abc


class CustomerFrontend(abc.ABC):
    """This component is responsible for interacting with the customers
    within some external system (e.g. Telegram)."""

    async def initialize(self):
        """Performs asynchronous initialization if needed."""

    @abc.abstractmethod
    async def start(self):
        """Starts serving the frontend, be it long polling loop or an HTTP server."""


class AgentFrontend(abc.ABC):
    async def initialize(self):
        pass

    @abc.abstractmethod
    async def start(self):
        pass


class ManagerFrontend(abc.ABC):
    async def initialize(self):
        pass

    @abc.abstractmethod
    async def start(self):
        pass



"""Support images

Revision ID: 5fc7b901794b
Revises: 48f9a7635310
Create Date: 2025-03-08 14:16:07.810681

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5fc7b901794b"
down_revision = "48f9a7635310"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "suppgram_events",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column(
            "kind",
            sa.Enum(
                "AGENT_ASSIGNED",
                "CONVERSATION_POSTPONED",
                "CONVERSATION_RATED",
                "CONVERSATION_RESOLVED",
                "CONVERSATION_STARTED",
                "CONVERSATION_TAG_ADDED",
                "CONVERSATION_TAG_REMOVED",
                "MESSAGE_SENT",
                name="eventkind",
            ),
            nullable=False,
        ),
        sa.Column("time_utc", sa.DateTime(timezone=True), nullable=False),
        sa.Column("agent_id", sa.Integer(), nullable=True),
        sa.Column("conversation_id", sa.Integer(), nullable=True),
        sa.Column("customer_id", sa.Integer(), nullable=True),
        sa.Column(
            "message_kind",
            sa.Enum("FROM_CUSTOMER", "FROM_AGENT", "POSTPONED", "RESOLVED", name="messagekind"),
            nullable=True,
        ),
        sa.Column(
            "message_media_kind",
            sa.Enum("NONE", "TEXT", "IMAGE", name="messagemediakind"),
            nullable=True,
        ),
        sa.Column("tag_id", sa.Integer(), nullable=True),
        sa.Column("workplace_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["agent_id"],
            ["suppgram_agents.id"],
        ),
        sa.ForeignKeyConstraint(
            ["conversation_id"],
            ["suppgram_conversations.id"],
        ),
        sa.ForeignKeyConstraint(
            ["customer_id"],
            ["suppgram_customers.id"],
        ),
        sa.ForeignKeyConstraint(
            ["tag_id"],
            ["suppgram_tags.id"],
        ),
        sa.ForeignKeyConstraint(
            ["workplace_id"],
            ["suppgram_workplaces.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.add_column(
        "suppgram_conversation_messages", sa.Column("image", sa.LargeBinary(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("suppgram_conversation_messages", "image")
    op.drop_table("suppgram_events")
    # ### end Alembic commands ###



from typing import List, Mapping

from telegram.ext import Application, ApplicationBuilder


class TelegramAppManager:
    def __init__(self, apps: List[Application]):
        self._apps: Mapping[str, Application] = {app.bot.token: app for app in apps}

    @classmethod
    def from_tokens(cls, tokens: List[str]):
        apps = [ApplicationBuilder().token(token).build() for token in tokens]
        return cls(apps)

    def get_app(self, token: str) -> Application:
        return self._apps[token]



import asyncio
from datetime import datetime, timezone
from typing import Optional
from uuid import uuid4

import aioconsole
import click
from pubnub.callbacks import SubscribeCallback
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.pubsub import PNMessageResult, PNPresenceEventResult
from pubnub.pubnub import PubNub

from suppgram.entities import MessageKind, Message, MessageMediaKind
from suppgram.frontends.pubnub.configuration import make_pubnub_configuration
from suppgram.frontends.pubnub.converter import (
    MessageConverter,
    make_pubnub_message_converter,
)


@click.command()
@click.option("--pubnub-user-id", default=None, help="PubNub user ID [default: random UUID]")
@click.option(
    "--pubnub-channel",
    default=None,
    help="PubNub channel for communication with support [default: UUID-support]",
)
@click.option(
    "--pubnub-channel-group",
    default="support",
    show_default=True,
    help="PubNub channel group to add channel to",
)
@click.option(
    "--pubnub-message-converter",
    "pubnub_message_converter_class_path",
    default="suppgram.frontends.pubnub.DefaultMessageConverter",
    show_default=True,
    help="Class converting messages between PubNub JSONs and suppgram Message objects",
)
def run_pubnub_customer_client(
    pubnub_user_id: Optional[str],
    pubnub_channel: Optional[str],
    pubnub_channel_group: str,
    pubnub_message_converter_class_path: str,
):
    loop = asyncio.get_event_loop()

    if not pubnub_user_id:
        pubnub_user_id = uuid4().hex
    if not pubnub_channel:
        pubnub_channel = f"{pubnub_user_id}-support"
    pubnub_configuration = make_pubnub_configuration(pubnub_user_id)
    pubnub = pubnub_configuration.instantiate_async()

    converter = make_pubnub_message_converter(pubnub_message_converter_class_path)

    async def _run():
        await pubnub.add_channel_to_channel_group().channels([pubnub_channel]).channel_group(
            pubnub_channel_group
        ).future()
        pubnub.add_listener(_SubscribeCallback(converter))
        pubnub.subscribe().channels([pubnub_channel]).execute()
        while True:
            text = await aioconsole.ainput("You: ")
            message = Message(
                kind=MessageKind.FROM_CUSTOMER,
                time_utc=datetime.now(timezone.utc),
                text=text,
            )
            converted = converter.convert_to_pubnub(message)
            await pubnub.publish().message(converted).channel(pubnub_channel).future()

    loop.run_until_complete(_run())


class _SubscribeCallback(SubscribeCallback):
    def __init__(self, message_converter: MessageConverter) -> None:
        self._message_converter = message_converter

    def status(self, pubnub: PubNub, status: PNStatus) -> None:
        pass

    def message(self, pubnub: PubNub, message: PNMessageResult) -> None:
        converted = self._message_converter.convert_from_pubnub(message)
        if converted.kind == MessageKind.FROM_CUSTOMER:
            return
        prefix = {
            MessageKind.FROM_AGENT: "\rAgent: ",
            MessageKind.RESOLVED: "\rConversation was marked as resolved.",
        }.get(converted.kind, f"\r{converted.kind.value}")
        text = converted.text or {MessageMediaKind.IMAGE: "[🖼️image]"}.get(converted.media_kind, "")
        print(f"{prefix}{text}\nYou: ", end="")

    def presence(self, pubnub: PubNub, presence: PNPresenceEventResult) -> None:
        pass


if __name__ == "__main__":
    run_pubnub_customer_client()



from pubnub.models.consumer.pubsub import PNMessageResult

from suppgram.entities import CustomerIdentification


def make_customer_identification(message: PNMessageResult) -> CustomerIdentification:
    return CustomerIdentification(pubnub_user_id=message.publisher, pubnub_channel_id=message.channel)



import abc
from dataclasses import dataclass
from enum import Enum
from typing import FrozenSet, Any, Optional, List

from suppgram.entities import CustomerIdentification


class TelegramChatRole(int, Enum):
    # Group for notifications on new unassigned conversations.
    NEW_CONVERSATION_NOTIFICATIONS = 1
    # Group all members of which are agents.
    AGENTS = 2


@dataclass(frozen=True)
class TelegramChat:
    telegram_chat_id: int
    roles: FrozenSet[TelegramChatRole]


class TelegramMessageKind(str, Enum):
    NEW_CONVERSATION_NOTIFICATION = "new_conversation_notification"
    RATE_CONVERSATION = "rate_conversation"
    NUDGE_TO_START_BOT_NOTIFICATION = "nudge_to_start_bot"
    CUSTOMER_MESSAGE_HISTORY = "customer_message_history"


@dataclass(frozen=True)
class TelegramMessage:
    id: Any

    telegram_bot_id: int
    chat: TelegramChat
    telegram_message_id: int
    kind: TelegramMessageKind

    # Here be all possible parameters of all kinds of messages:
    agent_id: Optional[Any]
    customer_id: Optional[Any]
    conversation_id: Optional[Any]
    telegram_bot_username: Optional[str]

    @property
    def customer_identification(self) -> CustomerIdentification:
        if self.customer_id is None:
            raise ValueError(
                f"no customer ID in message {self.telegram_message_id} in group {self.chat.telegram_chat_id}"
            )
        return CustomerIdentification(id=self.customer_id)


class TelegramStorage(abc.ABC):
    """Persistent storage for data specific to Telegram frontend.

    Currently, two entities are stored: Telegram chats and messages within chats.
    We need these data to track group roles and edit messages sent by bots if needed."""

    async def initialize(self):
        pass

    @abc.abstractmethod
    async def get_chat(self, telegram_chat_id: int) -> TelegramChat:
        """Fetch Telegram chat by Telegram chat ID."""

    @abc.abstractmethod
    async def create_or_update_chat(self, telegram_chat_id: int) -> TelegramChat:
        """Create or update Telegram chat by Telegram chat ID."""

    @abc.abstractmethod
    async def add_chat_roles(self, telegram_chat_id: int, *roles: TelegramChatRole) -> object:
        """Assign roles to a Telegram chat."""

    @abc.abstractmethod
    async def get_chats_by_role(self, role: TelegramChatRole) -> List[TelegramChat]:
        """Fetch all Telegram chats which have been assigned a role."""

    @abc.abstractmethod
    async def insert_message(
        self,
        telegram_bot_id: int,
        chat: TelegramChat,
        telegram_message_id: int,
        kind: TelegramMessageKind,
        *,
        agent_id: Optional[Any] = None,
        customer_id: Optional[Any] = None,
        conversation_id: Optional[Any] = None,
        telegram_bot_username: Optional[str] = None,
    ) -> TelegramMessage:
        """Store information about a Telegram message."""

    @abc.abstractmethod
    async def get_message(self, chat: TelegramChat, telegram_message_id: int) -> TelegramMessage:
        """Fetch a Telegram message."""

    @abc.abstractmethod
    async def get_messages(
        self,
        kind: TelegramMessageKind,
        *,
        agent_id: Optional[Any] = None,
        conversation_id: Optional[Any] = None,
        telegram_bot_username: Optional[str] = None,
    ) -> List[TelegramMessage]:
        """Fetch all Telegram messages satisfying condition(s)."""

    @abc.abstractmethod
    async def delete_messages(self, messages: List[TelegramMessage]):
        """Delete given Telegram messages."""

    @abc.abstractmethod
    async def get_newer_messages_of_kind(
        self, messages: List[TelegramMessage]
    ) -> List[TelegramMessage]:
        """For all given messages, find all newer messages in the corresponding chats.

        "Newer" here means "with greater Telegram message ID"."""



from telegram import Update, User

from suppgram.entities import (
    WorkplaceIdentification,
    AgentIdentification,
    CustomerIdentification,
    AgentDiff,
    CustomerDiff,
)


def make_agent_identification(effective_user: User) -> AgentIdentification:
    return AgentIdentification(telegram_user_id=effective_user.id)


def make_agent_diff(effective_user: User) -> AgentDiff:
    return AgentDiff(
        telegram_first_name=effective_user.first_name,
        telegram_last_name=effective_user.last_name,
        telegram_username=effective_user.username,
    )


def make_customer_identification(effective_user: User) -> CustomerIdentification:
    return CustomerIdentification(telegram_user_id=effective_user.id)


def make_customer_diff(effective_user: User) -> CustomerDiff:
    return CustomerDiff(
        telegram_first_name=effective_user.first_name,
        telegram_last_name=effective_user.last_name,
        telegram_username=effective_user.username,
    )


def make_workplace_identification(
    update: Update, effective_user: User
) -> WorkplaceIdentification:
    return WorkplaceIdentification(
        telegram_user_id=effective_user.id,
        telegram_bot_id=update.get_bot().id,
    )



from dataclasses import replace
from datetime import datetime, timezone
from typing import List, Any, Optional, Dict, AsyncIterator
from uuid import uuid4

from suppgram.containers import UnavailableList
from suppgram.entities import (
    Conversation,
    Message,
    Agent,
    Customer,
    WorkplaceIdentification,
    ConversationDiff,
    Tag,
    Workplace,
    AgentIdentification,
    AgentDiff,
    CustomerIdentification,
    CustomerDiff,
    ConversationState,
    SetNone,
    Event,
)
from suppgram.errors import (
    AgentNotFound,
    TagAlreadyExists,
    WorkplaceNotFound,
    ConversationNotFound,
    ConversationAlreadyAssigned,
)
from suppgram.storage import Storage


class InMemoryStorage(Storage):
    """In-memory implementation of [Storage][suppgram.storage.Storage] used in tests."""

    def __init__(self) -> None:
        self.customers: List[Customer] = []
        self.agents: List[Agent] = []
        self.workplaces: List[Workplace] = []
        self.tags: List[Tag] = []
        self.conversations: List[Conversation] = []
        self.events: List[Event] = []

    async def create_or_update_customer(
        self, identification: CustomerIdentification, diff: Optional[CustomerDiff] = None
    ) -> Customer:
        try:
            idx = next(
                i for i, c in enumerate(self.customers) if self._match_customer(identification, c)
            )
            customer = self.customers.pop(idx)
        except StopIteration:
            customer = self._construct_customer(identification)
        customer = self._update_customer(customer, diff)
        self.customers.append(customer)
        return customer

    async def get_agent(self, identification: AgentIdentification) -> Agent:
        try:
            return next(a for a in self.agents if self._match_agent(identification, a))
        except StopIteration:
            raise AgentNotFound(identification)

    async def find_all_agents(self) -> AsyncIterator[Agent]:
        for agent in self.agents:
            yield agent

    async def create_or_update_agent(
        self, identification: AgentIdentification, diff: Optional[AgentDiff] = None
    ) -> Agent:
        try:
            idx = next(i for i, a in enumerate(self.agents) if self._match_agent(identification, a))
            agent = self.agents.pop(idx)
        except StopIteration:
            agent = self._construct_agent(identification)
        agent = self._update_agent(agent, diff)
        self.agents.append(agent)
        return agent

    async def update_agent(self, identification: AgentIdentification, diff: AgentDiff) -> Agent:
        try:
            idx, agent = next(
                (i, a) for i, a in enumerate(self.agents) if self._match_agent(identification, a)
            )
        except StopIteration:
            raise AgentNotFound(identification)
        self.agents[idx] = agent = self._update_agent(agent, diff)
        return agent

    async def get_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        try:
            return next(w for w in self.workplaces if self._match_workplace(identification, w))
        except StopIteration:
            raise WorkplaceNotFound(identification)

    async def get_agent_workplaces(self, agent: Agent) -> List[Workplace]:
        return [w for w in self.workplaces if w.agent.id == agent.id]

    async def get_or_create_workplace(self, identification: WorkplaceIdentification) -> Workplace:
        try:
            return next(w for w in self.workplaces if self._match_workplace(identification, w))
        except StopIteration:
            agent = await self.get_agent(identification.to_agent_identification())
            workplace = self._construct_workplace(identification, agent)
            self.workplaces.append(workplace)
            return workplace

    async def create_tag(self, name: str, created_by: Agent) -> Tag:
        if any(t.name == name for t in self.tags):
            raise TagAlreadyExists(name)
        tag = Tag(
            id=name, name=name, created_at_utc=datetime.now(timezone.utc), created_by=created_by
        )
        self.tags.append(tag)
        return tag

    async def find_all_tags(self) -> List[Tag]:
        return [*self.tags]

    async def get_or_create_conversation(self, customer: Customer) -> Conversation:
        try:
            conversation = next(
                c
                for c in self.conversations
                if c.customer.id == customer.id and c.state != ConversationState.RESOLVED
            )
        except StopIteration:
            conversation = Conversation(
                id=uuid4().hex,
                state=ConversationState.NEW,
                customer=customer,
                tags=[],
            )
            self.conversations.append(conversation)
        return conversation

    async def find_conversations_by_ids(
        self, conversation_ids: List[Any], with_messages: bool = False
    ) -> List[Conversation]:
        return [
            c if with_messages else self._strip_messages(c)
            for c in self.conversations
            if c.id in conversation_ids
        ]

    async def find_all_conversations(
        self, with_messages: bool = False
    ) -> AsyncIterator[Conversation]:
        for c in self.conversations:
            yield c if with_messages else self._strip_messages(c)

    async def count_all_conversations(self) -> int:
        return len(self.conversations)

    async def update_conversation(
        self, id: Any, diff: ConversationDiff, unassigned_only: bool = False
    ):
        try:
            idx = next(i for i, c in enumerate(self.conversations) if c.id == id)
        except StopIteration:
            raise ConversationNotFound()

        if unassigned_only and self.conversations[idx].assigned_agent is not None:
            raise ConversationAlreadyAssigned()

        conv = self.conversations.pop(idx)

        changes: Dict[str, Any] = {}
        if diff.state is not None:
            changes["state"] = diff.state
        if diff.assigned_workplace_id is SetNone:
            changes["assigned_workplace"] = None
            changes["assigned_agent"] = None
        elif diff.assigned_workplace_id is not None:
            workplace = await self.get_workplace(
                WorkplaceIdentification(id=diff.assigned_workplace_id)
            )
            changes["assigned_workplace"] = workplace
            changes["assigned_agent"] = workplace.agent
        if diff.customer_rating is not None:
            changes["customer_rating"] = diff.customer_rating
        if diff.added_tags or diff.removed_tags:
            added_tags = diff.added_tags or []
            removed_tag_names = {tag.name for tag in diff.removed_tags or []}
            filtered_tags = [tag for tag in conv.tags if tag.name not in removed_tag_names]
            changes["tags"] = [*filtered_tags, *added_tags]

        conv = replace(conv, **changes)
        self.conversations.append(conv)
        return conv

    async def get_agent_conversation(self, identification: WorkplaceIdentification) -> Conversation:
        try:
            return next(
                c
                for c in self.conversations
                if c.assigned_workplace
                and self._match_workplace(identification, c.assigned_workplace)
            )
        except StopIteration:
            raise ConversationNotFound()

    async def find_customer_conversations(
        self, customer: Customer, with_messages: bool = False
    ) -> List[Conversation]:
        return [
            c if with_messages else self._strip_messages(c)
            for c in self.conversations
            if c.customer.id == customer.id
        ]

    async def find_agent_conversations(
        self, agent: Agent, with_messages: bool = False
    ) -> List[Conversation]:
        return [
            c if with_messages else self._strip_messages(c)
            for c in self.conversations
            if c.assigned_agent and c.assigned_agent.id == agent.id
        ]

    async def save_message(self, conversation: Conversation, message: Message):
        try:
            idx = next(i for i, c in enumerate(self.conversations) if c.id == conversation.id)
            conv = self.conversations.pop(idx)
        except StopIteration:
            raise ConversationNotFound()
        conv = replace(conv, messages=[*conv.messages, message])
        self.conversations.append(conv)

    async def save_event(self, event: Event):
        self.events.append(event)

    async def find_all_events(self) -> AsyncIterator[Event]:
        for event in self.events:
            yield event

    async def count_all_events(self) -> int:
        return len(self.events)

    def _match_customer(self, identification: CustomerIdentification, customer: Customer) -> bool:
        return (
            (customer.id == identification.id is not None)
            or (customer.telegram_user_id == identification.telegram_user_id is not None)
            or (customer.shell_uuid == identification.shell_uuid is not None)
            or (
                customer.pubnub_user_id == identification.pubnub_user_id is not None
                and customer.pubnub_channel_id == identification.pubnub_channel_id is not None
            )
        )

    def _construct_customer(self, identification: CustomerIdentification) -> Customer:
        if identification.id is not None:
            raise ValueError("can't create customer with predefined ID")
        return Customer(
            id=uuid4().hex,
            telegram_user_id=identification.telegram_user_id,
            shell_uuid=identification.shell_uuid,
            pubnub_user_id=identification.pubnub_user_id,
            pubnub_channel_id=identification.pubnub_channel_id,
        )

    def _update_customer(self, customer: Customer, diff: Optional[CustomerDiff]) -> Customer:
        if diff is None:
            return customer
        changes: Dict[str, Any] = {
            "telegram_first_name": diff.telegram_first_name,
            "telegram_last_name": diff.telegram_last_name,
            "telegram_username": diff.telegram_username,
        }
        changes = {k: v for k, v in changes.items() if v is not None}
        return replace(customer, **changes)

    def _match_agent(self, identification: AgentIdentification, agent: Agent) -> bool:
        return (agent.id == identification.id is not None) or (
            agent.telegram_user_id == identification.telegram_user_id is not None
        )

    def _construct_agent(self, identification: AgentIdentification) -> Agent:
        if identification.id is not None:
            raise ValueError("can't create agent with predefined ID")
        return Agent(
            id=uuid4().hex,
            deactivated=False,
            telegram_user_id=identification.telegram_user_id,
        )

    def _update_agent(self, agent: Agent, diff: Optional[AgentDiff]) -> Agent:
        if diff is None:
            return agent
        changes: dict[str, Any] = {
            "deactivated": diff.deactivated,
            "telegram_first_name": diff.telegram_first_name,
            "telegram_last_name": diff.telegram_last_name,
            "telegram_username": diff.telegram_username,
        }
        changes = {k: v for k, v in changes.items() if v is not None}
        return replace(agent, **changes)

    def _match_workplace(
        self, identification: WorkplaceIdentification, workplace: Workplace
    ) -> bool:
        return (workplace.id == identification.id is not None) or (
            workplace.telegram_user_id == identification.telegram_user_id is not None
            and workplace.telegram_bot_id == identification.telegram_bot_id is not None
        )

    def _construct_workplace(
        self, identification: WorkplaceIdentification, agent: Agent
    ) -> Workplace:
        if identification.id is not None:
            raise ValueError("can't create workplace with predefined ID")
        return Workplace(
            id=uuid4().hex,
            telegram_user_id=identification.telegram_user_id,
            telegram_bot_id=identification.telegram_bot_id,
            agent=agent,
        )

    def _strip_messages(self, conv: Conversation) -> Conversation:
        return replace(conv, messages=UnavailableList[Message]())



from datetime import timezone, datetime
from typing import (
    Any,
    Mapping,
    Optional,
    List,
    Union,
    Set,
    Iterable,
    NamedTuple,
    Dict,
)

from bson import ObjectId, CodecOptions, UuidRepresentation
from motor.core import AgnosticDatabase

from suppgram.containers import UnavailableList
from suppgram.entities import (
    Agent,
    AgentIdentification,
    AgentDiff,
    WorkplaceIdentification,
    Workplace,
    Tag,
    CustomerIdentification,
    CustomerDiff,
    Customer,
    FINAL_STATES,
    ConversationState,
    Conversation,
    Message,
    MessageKind,
    ConversationDiff,
    SetNone,
    Event,
    MessageMediaKind,
    EventKind,
)
from suppgram.errors import (
    AgentEmptyIdentification,
    WorkplaceEmptyIdentification,
    WorkplaceNotFound,
    CustomerEmptyIdentification,
)

Document = Mapping[str, Any]


class ConversationRelatedIDs(NamedTuple):
    customer_id: Any
    assigned_agent_id: Any
    assigned_workplace_id: Any

    @property
    def customer_identification(self) -> CustomerIdentification:
        return CustomerIdentification(id=self.customer_id)

    @property
    def assigned_workplace_identification(self) -> Optional[WorkplaceIdentification]:
        if self.assigned_workplace_id is None:
            return None
        return WorkplaceIdentification(id=self.assigned_workplace_id)


class Collections:
    """Abstraction layer over MongoDB database and collection names and BSON documents structure."""

    def __init__(
        self,
        database: AgnosticDatabase,
        customer_collection_name: str = "suppgram_customers",
        agent_collection_name: str = "suppgram_agents",
        conversation_collection_name: str = "suppgram_conversations",
        tag_collection_name: str = "suppgram_tags",
        event_collection_name: str = "suppgram_events",
        codec_options: CodecOptions = CodecOptions(
            tz_aware=True, tzinfo=timezone.utc, uuid_representation=UuidRepresentation.STANDARD
        ),
    ):
        """
        Parameters:
            database: asyncio-compatible Motor MongoDB database to store data in
            customer_collection_name: name of MongoDB collection to store customers in
            agent_collection_name: name of MongoDB collection to store agents and workplaces in
            conversation_collection_name: name of MongoDB collection to store conversations and messages in
            tag_collection_name: name of MongoDB collection to store tags in
        """
        self.customer_collection = database.get_collection(customer_collection_name, codec_options)
        self.agent_collection = database.get_collection(agent_collection_name, codec_options)
        # Workplaces are stored within agents.
        self.conversation_collection = database.get_collection(
            conversation_collection_name, codec_options
        )
        # Messages are stored within conversations.
        self.tag_collection = database.get_collection(tag_collection_name, codec_options)
        self.event_collection = database.get_collection(event_collection_name, codec_options)

    def make_customer_filter(self, identification: CustomerIdentification) -> Document:
        if identification.id is not None:
            return {"_id": ObjectId(identification.id)}
        if identification.telegram_user_id is not None:
            return {"telegram_user_id": identification.telegram_user_id}
        if identification.pubnub_user_id is not None:
            return {
                "pubnub_user_id": identification.pubnub_user_id,
                "pubnub_channel_id": identification.pubnub_channel_id,
            }
        if identification.shell_uuid is not None:
            return {"shell_uuid": identification.shell_uuid}  # MongoDB supports UUIDs
        raise CustomerEmptyIdentification(identification)

    def make_customers_filter(self, ids: Iterable[Any]) -> Document:
        return {"_id": {"$in": [ObjectId(id) for id in ids]}}

    def convert_to_customer(self, doc: Document) -> Customer:
        return Customer(
            id=str(doc["_id"]),
            telegram_user_id=doc.get("telegram_user_id"),
            telegram_first_name=doc.get("telegram_first_name"),
            telegram_last_name=doc.get("telegram_last_name"),
            telegram_username=doc.get("telegram_username"),
            shell_uuid=doc.get("shell_uuid"),  # MongoDB supports UUIDs
            pubnub_user_id=doc.get("pubnub_user_id"),
            pubnub_channel_id=doc.get("pubnub_channel_id"),
        )

    def make_customer_update(
        self, identification: CustomerIdentification, diff: Optional[CustomerDiff]
    ) -> Document:
        result: Dict[str, Any] = {}
        if identification.telegram_user_id is not None:
            result["telegram_user_id"] = identification.telegram_user_id
        if identification.pubnub_user_id is not None:
            result["pubnub_user_id"] = identification.pubnub_user_id
            result["pubnub_channel_id"] = identification.pubnub_channel_id
        if identification.shell_uuid is not None:
            result["shell_uuid"] = identification.shell_uuid  # MongoDB supports UUIDs
        if diff is None:
            return {"$set": result}
        if diff.telegram_first_name is not None:
            result["telegram_first_name"] = diff.telegram_first_name
        if diff.telegram_last_name is not None:
            result["telegram_last_name"] = diff.telegram_last_name
        if diff.telegram_username is not None:
            result["telegram_username"] = diff.telegram_username
        return {"$set": result}

    def make_agent_filter(
        self, identification: Union[AgentIdentification, WorkplaceIdentification]
    ) -> Document:
        if isinstance(identification, AgentIdentification) and identification.id is not None:
            return {"_id": ObjectId(identification.id)}
        if isinstance(identification, WorkplaceIdentification) and identification.id is not None:
            return {"workplaces.id": identification.id}
        if identification.telegram_user_id is not None:
            return {"telegram_user_id": identification.telegram_user_id}
        raise AgentEmptyIdentification(
            identification.to_agent_identification()
            if isinstance(identification, WorkplaceIdentification)
            else identification
        )

    def make_agents_filter(self, agent_ids: Iterable[Any]) -> Document:
        return {"_id": {"$in": [ObjectId(id) for id in agent_ids]}}

    def make_agents_filter_by_workplace_ids(self, workplace_ids: Iterable[Any]) -> Document:
        return {"workplaces.id": {"$in": list(workplace_ids)}}

    def convert_to_agent(self, agent_doc: Document) -> Agent:
        return Agent(
            id=str(agent_doc["_id"]),
            deactivated=bool(agent_doc.get("deactivated")),
            telegram_user_id=agent_doc.get("telegram_user_id"),
            telegram_first_name=agent_doc.get("telegram_first_name"),
            telegram_last_name=agent_doc.get("telegram_last_name"),
            telegram_username=agent_doc.get("telegram_username"),
        )

    def convert_to_agent_update(
        self, identification: AgentIdentification, diff: Optional[AgentDiff]
    ) -> Document:
        result: Dict[str, Any] = {"$set": {}, "$setOnInsert": {"workplaces": []}}
        if identification.telegram_user_id is not None:
            result["$set"]["telegram_user_id"] = identification.telegram_user_id
        if diff is None:
            return result
        if diff.deactivated is not None:
            result["$set"]["deactivated"] = diff.deactivated
        if diff.telegram_first_name is not None:
            result["$set"]["telegram_first_name"] = diff.telegram_first_name
        if diff.telegram_last_name is not None:
            result["$set"]["telegram_last_name"] = diff.telegram_last_name
        if diff.telegram_username is not None:
            result["$set"]["telegram_username"] = diff.telegram_username
        return result

    def convert_to_workplace(
        self, identification: WorkplaceIdentification, agent_doc: Document
    ) -> Workplace:
        workplace_subdocs = agent_doc["workplaces"]
        if identification.id is not None:

            def predicate(subdoc: Document):
                return subdoc["id"] == identification.id

        elif identification.telegram_bot_id is not None:

            def predicate(subdoc: Document):
                return subdoc["telegram_bot_id"] == identification.telegram_bot_id

        else:
            raise WorkplaceEmptyIdentification(identification)
        try:
            workplace_doc = next(
                workplace_doc for workplace_doc in workplace_subdocs if predicate(workplace_doc)
            )
        except StopIteration:
            raise WorkplaceNotFound(identification)
        return self._convert_from_workplace_subdocument(agent_doc, workplace_doc)

    def convert_to_workplaces(self, agent_doc: Document) -> List[Workplace]:
        workplace_subdocs = agent_doc["workplaces"]
        return [
            self._convert_from_workplace_subdocument(agent_doc, workplace_subdoc)
            for workplace_subdoc in workplace_subdocs
        ]

    def _convert_from_workplace_subdocument(
        self, agent_doc: Document, workplace_doc: Document
    ) -> Workplace:
        return Workplace(
            id=str(workplace_doc["id"]),
            telegram_user_id=agent_doc.get("telegram_user_id"),
            telegram_bot_id=workplace_doc["telegram_bot_id"],
            agent=self.convert_to_agent(agent_doc),
        )

    def convert_to_workspace_update(
        self, agent_id: Any, identification: WorkplaceIdentification
    ) -> Document:
        result: Dict[str, Any] = {}
        if identification.telegram_bot_id is not None:
            result["id"] = f"{agent_id}_{identification.telegram_bot_id}"
            result["telegram_bot_id"] = identification.telegram_bot_id
        else:
            raise WorkplaceEmptyIdentification(identification)
        # WARNING: once we add auxiliary fields to Workplace, $addToSet will stop being a viable solution.
        # Maybe we'll have to store metadata in a parallel array.
        return {"$addToSet": {"workplaces": result}}

    def convert_to_tag_document(self, name: str, created_by: Agent) -> Document:
        return {
            "_id": name,
            "created_at_utc": datetime.now(timezone.utc),
            "created_by": ObjectId(created_by.id),
        }

    def extract_agent_ids(self, tag_docs: List[Document]) -> Set[ObjectId]:
        return {tag_doc["created_by"] for tag_doc in tag_docs}

    def convert_to_tag(self, doc: Document, agents: Mapping[Any, Agent]) -> Tag:
        agent = agents[str(doc["created_by"])]
        return Tag(
            id=doc["_id"],
            name=doc["_id"],
            created_at_utc=doc["created_at_utc"],
            created_by=agent,
        )

    def make_conversation_filter_by_customer(self, customer: Customer) -> Document:
        return {"customer_id": ObjectId(customer.id), "state": {"$nin": FINAL_STATES}}

    def make_conversation_filter(self, id: Any, unassigned_only: bool) -> Document:
        result: Dict[str, Any] = {"_id": ObjectId(id)}
        if unassigned_only:
            result["assigned_workplace_id"] = None
        return result

    def make_conversations_filter(self, ids: Iterable[Any]) -> Document:
        return {"_id": {"$in": [ObjectId(id) for id in ids]}}

    def make_customer_conversations_filter(self, customer_id: Any) -> Document:
        return {"customer_id": ObjectId(customer_id)}

    def make_agent_conversations_filter(self, agent_id: Any) -> Document:
        return {"assigned_agent_id": ObjectId(agent_id)}

    def make_workplace_conversation_filter(self, workplace_id: Any) -> Document:
        return {"assigned_workplace_id": workplace_id}

    def make_conversation_update(
        self,
        customer: Optional[Customer] = None,
        diff: Optional[ConversationDiff] = None,
        workplaces: Mapping[Any, Workplace] = {},
    ) -> Document:
        result: Dict[str, Any] = {"$set": {}, "$setOnInsert": {}}
        if customer is not None:
            result["$setOnInsert"].update(
                {
                    "state": ConversationState.NEW,
                    "customer_id": ObjectId(customer.id),
                    "tag_ids": [],
                    "messages": [],
                }
            )
        if diff is not None:
            if diff.state is not None:
                result["$set"]["state"] = diff.state

            if diff.assigned_workplace_id is SetNone:
                result["$unset"] = {"assigned_agent_id": True, "assigned_workplace_id": True}
            elif diff.assigned_workplace_id is not None:
                result["$set"].update(
                    {
                        "assigned_agent_id": ObjectId(
                            workplaces[diff.assigned_workplace_id].agent.id
                        ),
                        "assigned_workplace_id": diff.assigned_workplace_id,
                    }
                )

            if diff.added_tags:
                result["$addToSet"] = {"tag_ids": {"$each": [tag.id for tag in diff.added_tags]}}

            if diff.removed_tags:
                result["$pull"] = {"tag_ids": {"$in": [tag.id for tag in diff.removed_tags]}}

            # TODO probably should allow adding and removing at the same time, which MongoDB forbids

            if diff.customer_rating is not None:
                result["$set"]["customer_rating"] = diff.customer_rating

        return result

    def extract_conversation_related_ids(self, conv_doc: Document) -> ConversationRelatedIDs:
        assigned_agent_id = conv_doc.get("assigned_agent_id")
        return ConversationRelatedIDs(
            customer_id=str(conv_doc["customer_id"]),
            assigned_agent_id=str(assigned_agent_id) if assigned_agent_id is not None else None,
            assigned_workplace_id=conv_doc.get("assigned_workplace_id"),
        )

    def make_conversation_projection(self, with_messages: bool) -> Document:
        if not with_messages:
            return {"messages": False}
        return {}

    def convert_to_conversation(
        self,
        doc: Document,
        customers: Mapping[Any, Customer],
        workplaces: Mapping[Any, Workplace],
        tags: Mapping[Any, Tag],
    ) -> Conversation:
        assigned_workplace_id = doc.get("assigned_workplace_id")
        assigned_workplace = (
            workplaces[str(assigned_workplace_id)] if assigned_workplace_id is not None else None
        )
        assigned_agent = assigned_workplace.agent if assigned_workplace is not None else None
        return Conversation(
            id=str(doc["_id"]),
            state=ConversationState(doc["state"]),
            customer=customers[str(doc["customer_id"])],
            tags=[tags[tag_id] for tag_id in doc["tag_ids"]],
            assigned_agent=assigned_agent,
            assigned_workplace=assigned_workplace,
            messages=[
                Message(
                    kind=MessageKind(message_doc["kind"]),
                    time_utc=message_doc["time_utc"],
                    text=message_doc.get("text"),
                    image=message_doc.get("image"),
                )
                for message_doc in doc["messages"]
            ]
            if "messages" in doc
            else UnavailableList[Message](),
            customer_rating=doc.get("customer_rating"),
        )

    def make_message_update(self, message: Message) -> Document:
        return {
            "$push": {
                "messages": {
                    "kind": message.kind,
                    "time_utc": message.time_utc,
                    **_maybe("text", message.text),
                    **_maybe("image", message.image),
                }
            }
        }

    def convert_to_event_document(self, event: Event) -> Document:
        doc = {
            "kind": event.kind,
            "time_utc": event.time_utc,
            "agent_id": ObjectId(event.agent_id),
            "conversation_id": ObjectId(event.conversation_id),
            "customer_id": ObjectId(event.customer_id),
            "message_kind": event.message_kind,
            "message_media_kind": event.message_media_kind,
            "tag_id": event.tag_id,
            "workplace_id": event.workplace_id,
        }
        return {k: v for k, v in doc.items() if v is not None}

    def convert_to_event(self, doc: Document) -> Event:
        return Event(
            kind=EventKind(doc["kind"]),
            time_utc=doc["time_utc"],
            agent_id=str(doc["agent_id"]) if "agent_id" in doc else None,
            conversation_id=str(doc["conversation_id"]) if "conversation_id" in doc else None,
            customer_id=str(doc["customer_id"]) if "customer_id" in doc else None,
            message_kind=MessageKind(doc["message_kind"]) if "message_kind" in doc else None,
            message_media_kind=MessageMediaKind(doc["message_media_kind"])
            if "message_media_kind" in doc
            else None,
            tag_id=doc.get("tag_id"),
            workplace_id=doc.get("workplace_id"),
        )


def _maybe(key: str, value: Any) -> dict[str, Any]:
    return {key: value} if value else {}



import abc
import asyncio
from typing import Callable, TypeVar, Generic, List, Awaitable

from suppgram.helpers import flat_gather

T = TypeVar("T")


class Observable(Generic[T], abc.ABC):
    @abc.abstractmethod
    def add_handler(
        self, handler: Callable[[T], Awaitable[None]]
    ) -> Callable[[T], Awaitable[None]]:
        pass

    @abc.abstractmethod
    def add_batch_handler(
        self, handler: Callable[[List[T]], Awaitable[None]]
    ) -> Callable[[List[T]], Awaitable[None]]:
        pass

    @abc.abstractmethod
    async def trigger(self, event: T):
        pass

    @abc.abstractmethod
    async def trigger_batch(self, events: List[T]):
        pass


class LocalObservable(Observable[T]):
    def __init__(self) -> None:
        self._handlers: List[Callable[[T], Awaitable[None]]] = []
        self._batch_handlers: List[Callable[[List[T]], Awaitable[None]]] = []

    def add_handler(
        self, handler: Callable[[T], Awaitable[None]]
    ) -> Callable[[T], Awaitable[None]]:
        self._handlers.append(handler)
        return handler  # return value can be used for decorator chaining

    def add_batch_handler(
        self, handler: Callable[[List[T]], Awaitable[None]]
    ) -> Callable[[List[T]], Awaitable[None]]:
        self._batch_handlers.append(handler)
        return handler  # return value can be used for decorator chaining

    async def trigger(self, event: T):
        await asyncio.gather(
            flat_gather(handler(event) for handler in self._handlers),
            flat_gather(handler([event]) for handler in self._batch_handlers),
        )

    async def trigger_batch(self, events: List[T]):
        await asyncio.gather(
            flat_gather(
                handler(event) for handler in self._handlers for event in events
            ),
            flat_gather(handler(events) for handler in self._batch_handlers),
        )



from typing import List, Optional

from telegram import Bot
from telegram.error import TelegramError

from suppgram.frontends.telegram import TelegramStorage
from suppgram.frontends.telegram.app_manager import TelegramAppManager
from suppgram.frontends.telegram.storage import TelegramChat, TelegramChatRole, TelegramMessage
from suppgram.helpers import flat_gather


class TelegramHelper:
    def __init__(
        self,
        manager_bot_token: Optional[str],
        app_manager: TelegramAppManager,
        storage: TelegramStorage,
    ):
        self._storage = storage
        self._manager_bot: Optional[Bot] = (
            app_manager.get_app(manager_bot_token).bot if manager_bot_token else None
        )

    async def check_belongs_to_agent_chats(self, telegram_user_id: int) -> List[TelegramChat]:
        chats = await self._storage.get_chats_by_role(TelegramChatRole.AGENTS)
        return [
            chat
            for chat in await flat_gather(
                self.check_belongs_to_group(telegram_user_id, group) for group in chats
            )
            if chat is not None
        ]

    async def check_belongs_to_group(
        self, telegram_user_id: int, group: TelegramChat
    ) -> Optional[TelegramChat]:
        if self._manager_bot is None:
            return None
        try:
            member = await self._manager_bot.get_chat_member(
                chat_id=group.telegram_chat_id, user_id=telegram_user_id
            )
            return (
                group
                if member.status in (member.ADMINISTRATOR, member.OWNER, member.MEMBER)
                else None
            )
        except TelegramError:  # TODO more precise exception handling
            return None

    async def delete_message_if_exists(self, message: TelegramMessage):
        if self._manager_bot is None:
            return
        try:
            await self._manager_bot.delete_message(
                chat_id=message.chat.telegram_chat_id,
                message_id=message.telegram_message_id,
            )
        except TelegramError:  # TODO more precise exception handling
            pass
